# Thanks to https://gist.github.com/dahlia/e618b303ab64c0e7e32b1f7957ca283a#file-wsgi_typing-py

import types
from typing import Any, Callable, Iterable, Mapping, Optional, Sequence, Tuple

__all__ = (
    "Environ",
    "ExcInfo",
    "ResponseStream",
    "StartResponse",
    "WriteCallable",
    "WsgiApplication",
)


# https://www.python.org/dev/peps/pep-3333/#the-write-callable
WriteCallable = Callable[[bytes], Any]

# https://docs.python.org/3/library/sys.html#sys.exc_info
ExcInfo = Tuple[type, BaseException, types.TracebackType]

# https://www.python.org/dev/peps/pep-3333/#the-start-response-callable
StartResponse = Callable[
    [
        str,  # status
        Sequence[Tuple[str, str]],  # response headers
        Optional[ExcInfo],  # exc_info
    ],
    WriteCallable,  # write() callable
]

# https://www.python.org/dev/peps/pep-3333/#environ-variables
Environ = Mapping[str, object]

# https://www.python.org/dev/peps/pep-3333/#buffering-and-streaming
ResponseStream = Iterable[bytes]

# https://www.python.org/dev/peps/pep-3333/#specification-details
WsgiApplication = Callable[
    [
        Environ,  # environ
        StartResponse,  # start_response() function
    ],
    ResponseStream,  # bytestring chunks
]



from threading import Lock
from time import sleep, monotonic
from math import ceil, inf
from os import environ

from psutil import cpu_count, cpu_times

from .cgroups import get_cgroups_cpu_quota, get_cpu_used

# How often to poll the CPU, compared to the moving-average of the job-start-interval
CPU_IDLE_POLL_DIVISOR = 20

# Polling the cpu-idle more often than this is bound to give noisy results
# If your application consumes jobs more often than this, pothead probably is not for you
CPU_IDLE_POLL_MIN_INTERVAL = 0.050
CPU_IDLE_POLL_MAX_INTERVAL = 2

# The worker will ignore available idle CPU until it is using at least this much CPU itself
# Useful for shared environments like in Kubernetes where neighbor pods might use your `cpu.request`
# oppotunistically, until you use it yourself
CPU_PROMISED = int(environ.get("POTHEAD_CPU_PROMISED", 0))

# Jobs will still be clamped to idle CPU, but we want SOME kind of upper bound on concurrency, to avoid
# temporary I/O blockage causing ridiculous amounts of work
DEFAULT_MAX_CONCURRENT_MULTIPLIER = 2

__all__ = ["wait_for_idle_cpus"]


class MovingAverage:
    def __init__(self, init=0, inertia=0.6):
        self.value = init
        self.speed = 1 - inertia
        self.inertia = inertia

    def update(self, value):
        self.value = (self.value * self.inertia) + (value * self.speed)
        return self.value


class Worker:
    def __init__(self, cpu_reservation_amount, keep_cpu_reservation):
        self.cpu_reservation_amount = cpu_reservation_amount
        self.keep_cpu_reservation = keep_cpu_reservation
        self.on_done = None

    def reset_cpu_reservation(self, new_cpu_amount=0):
        # Note: we rely on the GIL/atomicity of Python bytecode execution
        # here, since this value will be changed from the handler thread,
        # but read (via `reserved_cpu_amount()`) in the `wait_for_slot()`
        # thread.
        self.cpu_reservation_amount = new_cpu_amount
        # Note: we do not call on_done() here, even if the amount would be 0
        # since that would remove a live worker from the list, which would give
        # the wrong worker count

    # Worker callback implementation
    def request_received(self, environ):
        environ["RESET_CPU_RESERVATION"] = self.reset_cpu_reservation
        if not self.keep_cpu_reservation:
            self.reset_cpu_reservation(0)

    # Worker callback implementation
    def done(self):
        self.on_done(self)


class WorkerList:
    def __init__(self):
        self.workers = set()
        self.lock = Lock()

    def add_worker(self, worker):
        with self.lock:
            worker.on_done = lambda w: self.remove_worker(w)
            self.workers.add(worker)
        return worker

    def remove_worker(self, worker):
        with self.lock:
            self.workers.remove(worker)

    def __len__(self):
        with self.lock:
            return len(self.workers)

    def reserved_cpu_amount(self):
        with self.lock:
            return sum([w.cpu_reservation_amount for w in self.workers])


class InertialTimeDerivate:
    """Timed measurement on an incrementing counter

    Reads the counter, divides with time passed since last reading, and smooths the value using a moving average
    """

    def __init__(self, func, initial, inertia):
        self.func = func
        self.last_value = self.func()
        self.last_check = monotonic()
        self.inertial_value = MovingAverage(initial, inertia)

    def update(self):
        new_value = self.func()
        this_check = monotonic()
        elapsed = this_check - self.last_check
        delta = new_value - self.last_value
        self.last_check = this_check
        self.last_value = new_value
        return self.inertial_value.update(delta / elapsed)


def wait_for_idle_cpus(
    required, *, max_concurrent=None, inertia=0.7, keep_cpu_reservation=False
):
    """CPU-gated wait_for_slot implementation

    Creates a wait_for_slot-callable that will let through at least one concurrent job, and additional jobs
    as long as `required` cpu:s are idle. It is cgroup-aware, and will not fetch new jobs that would break
    configured cgroup-limit. It will also measure the total CPU _of the current cgroup_, as reported by
    /sys/fs/cgroup/cpu.stat. Typically, when running under Linux, in a container, or in the host operating
    system, this will do what's expected.

    This gater will accept new jobs as long as the current cgroup (container) aren't using
    `POTHEAD_CPU_PROMISED` cores, or if further idle CPU seems to be available.

    When a worker is spawned, but before it has received its job request, it will have the CPU amount
    specified in the `required` parameter artificially reserved for it. The `keep_cpu_reservation`
    parameter can be used for jobs that when started don't immediately use the full amount of CPU that
    they will do eventually.

    In order to avoid spawning lots of such jobs in a short time period before CPU usage has settled, the
    artifical CPU reservation can be kept by setting `keep_cpu_reservation` to True. When doing this, the job
     should at some later point during its request processing reset the
    CPU reservation by calling the function stored under the "RESET_CPU_RESERVATION" key in the WSGI
    request environment, like so:

        http_request.environ["RESET_CPU_RESERVATION"](new_cpu_amount)

    The function may be called several times during the request processing, but the last call should
    normally set the amount to 0.
    """
    worker_list = WorkerList()
    cpu_wait_time = MovingAverage(1, 0.7)

    cpu_quota = get_cgroups_cpu_quota()
    if cpu_quota < inf:
        if not max_concurrent:
            max_concurrent = ceil(
                cpu_quota * DEFAULT_MAX_CONCURRENT_MULTIPLIER / required
            )

        # Precalculate the quota with headroom for a new job
        cpu_quota -= required
    else:
        if not max_concurrent:
            max_concurrent = ceil(
                cpu_count() * DEFAULT_MAX_CONCURRENT_MULTIPLIER / required
            )

    def cpu_is_available(idle, cpu_used):
        reserved = worker_list.reserved_cpu_amount()
        cpu_used += reserved

        if cpu_used < CPU_PROMISED:
            return True

        if idle - reserved < required:
            return False

        return cpu_used <= cpu_quota

    def wait_for_slot(halt):
        start_time = monotonic()

        idle = InertialTimeDerivate(lambda: cpu_times().idle, 0, inertia)
        cpu_used = InertialTimeDerivate(get_cpu_used, cpu_count(), inertia)

        cpu_poll_interval = max(
            CPU_IDLE_POLL_MIN_INTERVAL,
            min(
                cpu_wait_time.value / CPU_IDLE_POLL_DIVISOR, CPU_IDLE_POLL_MAX_INTERVAL
            ),
        )
        while len(worker_list) and not halt():
            idle_deriv = idle.update()
            cpu_used_deriv = cpu_used.update()
            if len(worker_list) < max_concurrent and cpu_is_available(
                idle_deriv, cpu_used_deriv
            ):
                break

            sleep(cpu_poll_interval)

        cpu_wait_time.update(monotonic() - start_time)

        return worker_list.add_worker(Worker(required, keep_cpu_reservation))

    return wait_for_slot



from math import inf
from os.path import isfile
from multiprocessing import cpu_count

CGROUPS_FS_PREFIX = "/sys/fs/cgroup"

CGROUPS_V1_CPU_QUOTA = f"{CGROUPS_FS_PREFIX}/cpu/cpu.cfs_quota_us"
CGROUPS_V1_CPU_PERIOD = f"{CGROUPS_FS_PREFIX}/cpu/cpu.cfs_period_us"
CGROUPS_V1_CPUACCT_USAGE = f"{CGROUPS_FS_PREFIX}/cpu/cpuacct.usage"
CGROUPS_V2_CPU_MAX = f"{CGROUPS_FS_PREFIX}/cpu.max"
CGROUPS_V2_CPU_STAT = f"{CGROUPS_FS_PREFIX}/cpu.stat"


def get_cgroups_cpu_quota() -> float:
    if isfile(CGROUPS_V2_CPU_MAX):
        return get_cpu_quota_v2()
    elif isfile(CGROUPS_V1_CPU_QUOTA):
        return get_cpu_quota_v1()
    else:
        return cpu_count()


# Documentation: https://www.kernel.org/doc/html/latest/admin-guide/cgroup-v2.html
def get_cpu_quota_v2() -> float:
    with open(CGROUPS_V2_CPU_MAX, "rt") as cpu_max:
        cpu_max_line = cpu_max.read()
        return parse_cpu_quota_v2(cpu_max_line)


def get_cpu_used() -> float:
    if isfile(CGROUPS_V2_CPU_STAT):
        return get_cpu_used_v2()
    elif isfile(CGROUPS_V1_CPUACCT_USAGE):
        return get_cpu_used_v1()
    else:
        assert False, "Found no recognised cgroup to extract CPU for"


def get_cpu_used_v1():
    with open(CGROUPS_V1_CPUACCT_USAGE, "rt") as container_cpu_stat:
        return int(container_cpu_stat.read().strip()) / 1_000_000_000


def get_cpu_used_v2():
    with open(CGROUPS_V2_CPU_STAT, "rt") as container_cpu_stat:
        for line in container_cpu_stat.readlines():
            if line.startswith("usage_usec "):
                return int(line.strip().split()[1]) / 1_000_000


def parse_cpu_quota_v2(cpu_max_line: str) -> float:
    words = cpu_max_line.split()
    if len(words) != 2:
        raise Exception(
            f"Unexpected content in cgroups v2 cpu.max file: {cpu_max_line}"
        )

    (quota_str, period_str) = words
    if quota_str == "max":
        return inf

    quota = int(quota_str)
    period = int(period_str)
    if quota <= 0:
        raise Exception(
            f"Quota in cgroups v2 cpu.max file was <= 0. Content: {cpu_max_line}"
        )
    if period <= 0:
        raise Exception(
            f"Period in cgroups v2 cpu.max file was <= 0. Content: {cpu_max_line}"
        )

    return quota / period


# Documentation: https://www.kernel.org/doc/Documentation/scheduler/sched-bwc.txt
def get_cpu_quota_v1() -> float:
    with open(CGROUPS_V1_CPU_QUOTA, "rt") as quota, open(
        CGROUPS_V1_CPU_PERIOD, "rt"
    ) as period:
        quota_line = quota.read()
        period_line = period.read()
        return parse_cpu_quota_v1(quota_line, period_line)


def parse_cpu_quota_v1(quota_line: str, period_line: str) -> float:
    quota = int(quota_line)
    if quota > 0:
        period = int(period_line)
        return quota / period
    elif quota == 0:
        raise Exception("Quota for cgroups v1 was 0")
    else:
        # A quota of -1 means "no restriction"
        return inf



#!/usr/bin/env python3

from http import HTTPStatus
from logging import getLogger
from signal import SIGTERM, signal
from socket import (
    AF_INET,
    IPPROTO_TCP,
    SHUT_RDWR,
    SO_KEEPALIVE,
    SOCK_STREAM,
    SOL_SOCKET,
    TCP_KEEPCNT,
    TCP_KEEPIDLE,
    TCP_KEEPINTVL,
    getaddrinfo,
    socket,
)
from threading import Condition, Thread, Timer
from time import monotonic, sleep
from typing import Dict, Optional, Set
from uuid import uuid4
from wsgiref import simple_server

from werkzeug.serving import WSGIRequestHandler, BaseWSGIServer, make_server

from .gating import wait_for_idle_cpus
from .oob_response import OutOfBandResponder
from .resource_balancer import BoundedResourceBalancer, ResourceLoan
from .subprocess_middleware import SubprocessMiddleware
from .util import ObjectProxy, ThreadsTracker

LOGGER = getLogger("pothead.worker")
BROKER_CONNECT_RETRY_INTERVAL = 2


class DeferrableCleanup:
    def __init__(self, f):
        self.f = f

    def assume(self):
        f = self.f
        self.f = None
        return f

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, trace):
        if self.f:
            self.f()


class Flag:
    def __init__(self):
        self._set = False
        self._cond = Condition()

    def __bool__(self):
        return self._set

    def set(self):
        with self._cond:
            self._set = True
            self._cond.notify_all()

    def wait(self, timeout: float):
        with self._cond:
            self._cond.wait_for(lambda: self._set, timeout)


class Handler(WSGIRequestHandler):
    protocol_version = "HTTP/1.1"

    def __init__(
        self, socket, address, server, cb_request_received, cb_processing_done
    ):
        self.cb_request_received = cb_request_received
        self.env = None
        self.request_id = None
        self.request_received_at = None

        with DeferrableCleanup(cb_processing_done) as c:
            if isinstance(server.app, OutOfBandResponder):
                app = server.app
                server = ObjectProxy(server)
                server.app = lambda environ, start_response: app(
                    environ, start_response, c.assume
                )

            super().__init__(socket, address, server)

    def handle_one_request(self):
        try:
            # Handler may be used for several requests, so we must reset
            # environment overrides each time
            self.env = None
            super().handle_one_request()
        finally:
            self.close_connection = True

    def make_environ(self):
        environ = super().make_environ()

        if self.cb_request_received:
            self.cb_request_received(environ)

        return environ

    def parse_request(self):
        success = super().parse_request()
        if success:
            self.request_received_at = monotonic()
            self.request_id = self.headers.get("X-Request-Id", None) or str(uuid4())

        # This socket shall no longer be fast-closed on SIGTERM
        socket = self.request
        self.server.sockets_waiting_for_request.remove(socket)

        return success

    def log_message(self, format, *args):
        LOGGER.info("[%s] " + format, self.request_id, *args)

    def log_request(self, code="-", size="-"):
        if isinstance(code, HTTPStatus):
            code = code.value
        LOGGER.info(
            'Starting request [%s] "%s" %s %s',
            self.request_id,
            self.requestline,
            str(code),
            str(size),
        )

    def log_error(self, format, *args):
        LOGGER.error("[%s] " + format, self.request_id, *args)

    def log_elapsed(self):
        LOGGER.info(
            "[%s] Elapsed: %.3f",
            self.request_id,
            monotonic() - self.request_received_at,
        )


class LoadBalancer:
    REFRESH_INTERVAL = 5

    def __init__(self, host, port):
        self.host = host
        self.port = port
        self._balancer = BoundedResourceBalancer(1)
        self.refresh()

    def acquire_addr(self) -> ResourceLoan:
        return self._balancer.acquire()

    def refresh(self):
        try:
            results = getaddrinfo(self.host, self.port, type=SOCK_STREAM)
            hosts = (sockaddr for _f, _t, _p, _c, sockaddr in results)
            self._balancer.provision(hosts)
        except BaseException:
            LOGGER.exception(
                "Failed to refresh endpoints for %s:%d", self.host, self.port
            )
        t = Timer(self.REFRESH_INTERVAL, self.refresh)
        t.daemon = True
        t.start()


class Server:
    shutdown_signal = (
        False  # Must exist for werkzeug 1.x legacy. Can be dropped with werkzeug 2.x
    )
    ssl_context = None
    multithread = False
    multiprocess = False
    passthrough_errors = False
    inspection_server: Optional[BaseWSGIServer] = None

    def __init__(
        self,
        addr,
        app,
        *,
        load_balancer=LoadBalancer,
        redirect_port: Optional[int] = None,
        inspection_port: Optional[int] = None,
        use_request_subprocess: bool = False,
        oob_cleanup_interval_seconds: Optional[float] = None,
        oob_item_expiry_seconds: Optional[float] = None,
    ):
        (host, port) = addr
        # Set up base environment
        env: Dict[str, str] = {}
        env["SERVER_NAME"] = host
        env["GATEWAY_INTERFACE"] = "HTTP/1.1"
        env["SERVER_PORT"] = port
        env["REMOTE_HOST"] = ""
        env["CONTENT_LENGTH"] = ""
        env["SCRIPT_NAME"] = ""
        self.base_environ = env

        self.broker_balancer = load_balancer(host, port)
        self.sockets_waiting_for_request: Set[socket] = set()
        self.worker_tracker = ThreadsTracker()

        self.stopping = Flag()
        self.stopped = Flag()

        if inspection_port is not None:
            self.inspection_server = make_server(
                "", inspection_port, self.inspect, threaded=True
            )
            Thread(target=self.inspection_server.serve_forever, daemon=True).start()

        self.wait_for_slot = app.wait_for_slot
        del app.wait_for_slot
        self.app = app
        self.app = (
            SubprocessMiddleware(LOGGER, self.app)
            if use_request_subprocess
            else self.app
        )
        self.app = (
            OutOfBandResponder(
                self.app,
                redirect_port,
                oob_cleanup_interval_seconds,
                oob_item_expiry_seconds,
            )
            if redirect_port is not None
            else self.app
        )

    def inspect(self, environ, start_response):
        if environ.get("REQUEST_METHOD", None) != "GET":
            start_response("405 Only GET supported", ())
            return
        path = environ.get("PATH_INFO", None)
        if path == "/inspect":
            start_response("200 OK", ())
            yield f"{len(self.worker_tracker)} workers running\n".encode()
            yield f"{len(self.sockets_waiting_for_request)} blocked waiting for request\n".encode()
        elif path == "/apocalypse":
            start_response("200 OK", ())
            while not self.stopped.wait(5):
                yield b"Not yet\n"
            yield b"It's time\n"
        else:
            start_response("404 Not found", ())

    def worker(self):
        while not self.stopping:
            with self.worker_tracker, self.broker_balancer.acquire_addr() as addr:
                self._fetch_and_run_one_job(addr)

    def _fetch_and_run_one_job(
        self, addr, cb_request_received=None, cb_processing_done=lambda: None
    ):
        with socket(AF_INET, SOCK_STREAM) as s:
            self.sockets_waiting_for_request.add(s)
            try:
                s.connect(addr)
                self.server_address = s.getsockname()

                s.setsockopt(SOL_SOCKET, SO_KEEPALIVE, 1)
                s.setsockopt(IPPROTO_TCP, TCP_KEEPIDLE, 1)
                s.setsockopt(IPPROTO_TCP, TCP_KEEPINTVL, 2)
                s.setsockopt(IPPROTO_TCP, TCP_KEEPCNT, 3)
            except Exception as e:
                LOGGER.error("Couldn't connect to broker: %r", e)
                cb_processing_done()
                sleep(BROKER_CONNECT_RETRY_INTERVAL)
                return

            try:
                h = Handler(s, addr, self, cb_request_received, cb_processing_done)
            except BaseException:
                LOGGER.exception("In WSGI request handling")
                sleep(1)
                return

            try:
                s.shutdown(SHUT_RDWR)
            except OSError:
                # We shutdown to ensure flushing, and for courtesy, but we don't care about errors here
                pass

            if h.request_received_at:
                h.log_elapsed()

    def get_app(self):
        return self.app

    def shutdown(self):
        LOGGER.info("Shutting down workers...")
        self.stopping.set()
        for s in list(self.sockets_waiting_for_request):
            try:
                s.shutdown(SHUT_RDWR)
                s.close()
            except OSError:
                pass

        self.worker_tracker.drain()

        if isinstance(self.app, OutOfBandResponder):
            self.app.shutdown()
        LOGGER.info("Worker shutdown completed")
        self.stopped.set()
        if self.inspection_server is not None:
            self.inspection_server.shutdown()
        LOGGER.info("All services stopped")

    def poll_loop(self):
        """Uses app.wait_for_slot() spawn new workers dynamically. This is useful, for example to wait for
        enough CPU, memory, or disk space, to start next job.

        `wait_for_slot` is expected to block until it's time to poll for a new job. It is given a function
        `halt` which it should poll intermittently to check if this worker is undergoing `shutdown()`.

        It may return an object, if so `obj.request_received(environ)` will be called when a request is
        received and `obj.done()` when either the request is received, or if polling failed in which case
        `request_received()` was never called.

        See pothead.gating for some ready-made `wait_for_slot` implementations"""

        while not self.stopping:
            callbacks = self.wait_for_slot(lambda: bool(self.stopping))
            if self.stopping:
                break

            # This will block as long as we are waiting for requests from all broker-instances
            broker_instance_lease = self.broker_balancer.acquire_addr()
            if self.stopping:
                break

            Thread(
                target=self._run_one_poll, args=(broker_instance_lease, callbacks)
            ).start()

    def _run_one_poll(self, broker_instance_lease: ResourceLoan, callbacks):
        def request_received(environ):
            broker_instance_lease.release()
            getattr(callbacks, "request_received", lambda _: None)(environ)

        with self.worker_tracker, broker_instance_lease as addr:
            self._fetch_and_run_one_job(
                addr, request_received, getattr(callbacks, "done", lambda: None)
            )

    def log(self, type, msg, *args, **kwargs):
        getattr(LOGGER, type)(msg.rstrip(), *args, **kwargs)


def install_term_handler(f):
    previous = []

    def on_term(_signal, _stack):
        Thread(target=f).start()
        for p in previous:
            p()

    p = signal(SIGTERM, on_term)
    if p:
        previous.append(p)


demo_app = simple_server.demo_app
demo_app.wait_for_slot = wait_for_idle_cpus(3)


if __name__ == "__main__":
    from argparse import ArgumentParser
    from importlib import import_module
    from logging import INFO, basicConfig
    from os import environ
    from sys import path

    path.insert(0, ".")

    def address(str):
        (host, port) = str.rsplit(":", 1)
        return (host, int(port))

    def func(str):
        (module, symbol) = str.rsplit(":", 1)
        module = import_module(module)
        return getattr(module, symbol)

    DEFAULT_WORKERS = int(environ.get("POTHEAD_WORKERS", 1))

    parser = ArgumentParser(description="Run WSGI app in sequential `worker` mode")
    parser.add_argument(
        "--connect",
        default="localhost:4040",
        type=address,
        help="Load Balancer Hub to connect to [host:port]",
    )
    parser.add_argument(
        "--redirect-response",
        default=None,
        help="When streaming successful responses, first redirect client to worker at [port]",
    )
    parser.add_argument(
        "--inspection-port",
        default=None,
        type=int,
        help="If set, the server will serve a small inspection-API on [port]",
    )
    parser.add_argument(
        "--use-request-subprocess",
        action="store_true",
        help=(
            "Spawn a separate WSGI app subprocess for handling each request. "
            + "Note: Only suitable for use cases where the request payload is small."
        ),
    )

    job_control_group = parser.add_mutually_exclusive_group()
    job_control_group.add_argument(
        "--workers",
        default=DEFAULT_WORKERS,
        type=int,
        help="Number of worker Processes",
    )
    job_control_group.add_argument(
        "--poll-jobs",
        action="store_true",
        help="Use `app.wait_for_slot` to determine when to pull new jobs",
    )

    parser.add_argument(
        "app",
        nargs="?",
        default="pothead.worker:demo_app",
        type=func,
        help="The WSGI request handler to handle requests",
    )
    args = parser.parse_args()
    basicConfig(level=INFO)

    LOGGER.info("Initializing server for app %s", args.app)

    server = Server(
        args.connect,
        args.app,
        redirect_port=args.redirect_response,
        use_request_subprocess=args.use_request_subprocess,
        inspection_port=args.inspection_port,
    )

    install_term_handler(server.shutdown)

    if args.poll_jobs:
        LOGGER.info(
            "Starting up job-polling dynamic workers connecting to %s", args.connect
        )
        server.poll_loop()
        LOGGER.info("connect-loop shut down completed")
    else:
        LOGGER.info(
            "Starting up %d workers connecting to %s", args.workers, args.connect
        )
        workers = [Thread(target=server.worker) for _ in range(args.workers)]
        for worker in workers:
            worker.start()
        for worker in workers:
            worker.join()



from threading import get_ident, Condition
from typing import Optional, Set


class ObjectProxy:
    """Decorator allowing override of attributes on an instance of another object

    >>> s1 = "hello"
    >>> s2 = ObjectProxy(s1)
    >>> s2.encode = lambda: b"goodbye!"
    >>> assert s1.encode() == b"hello"
    >>> assert s2.encode() == b"goodbye!"
    """

    def __init__(self, inner):
        self.__inner = inner

    def __getattr__(self, name):
        return getattr(self.__inner, name)

    def __call__(self, *args, **kwargs):
        return self.__inner(*args, **kwargs)


class ThreadsTracker:
    _threads: Optional[Set[int]]

    def __init__(self):
        self._threads = set()
        self._cond = Condition()

    def __enter__(self):
        self._threads.add(get_ident())
        return self

    def __exit__(self, exc_type, exc_value, trace):
        with self._cond:
            self._threads.remove(get_ident())
            self._cond.notify_all()

    def __len__(self):
        with self._cond:
            return len(self._threads)

    def drain(self):
        with self._cond:
            self._cond.wait_for(lambda: len(self._threads) == 0)



from math import inf
from time import monotonic
from unittest import TestCase
from random import random

from .cgroups import parse_cpu_quota_v1, parse_cpu_quota_v2, get_cpu_used


class CgroupsTest(TestCase):
    def test_v1_valid(self):
        assert parse_cpu_quota_v1("500000", "500000") == 1.0
        assert parse_cpu_quota_v1("200000", "500000") == 0.4
        assert parse_cpu_quota_v1("-1", "500000") == inf

    def test_v1_invalid(self):
        self.verify_exception_with_msg(
            lambda: parse_cpu_quota_v1("0", "500000"), "Quota for cgroups v1 was 0"
        )

    def test_v2_valid(self):
        assert parse_cpu_quota_v2("500000 500000") == 1.0
        assert parse_cpu_quota_v2("200000 500000") == 0.4
        assert parse_cpu_quota_v2("max 500000") == inf

    def test_v2_invalid(self):
        self.verify_exception_with_msg(
            lambda: parse_cpu_quota_v2("0 500000"),
            "Quota in cgroups v2 cpu.max file was <= 0. Content: 0 500000",
        )
        self.verify_exception_with_msg(
            lambda: parse_cpu_quota_v2("-1 500000"),
            "Quota in cgroups v2 cpu.max file was <= 0. Content: -1 500000",
        )
        self.verify_exception_with_msg(
            lambda: parse_cpu_quota_v2("200000 0"),
            "Period in cgroups v2 cpu.max file was <= 0. Content: 200000 0",
        )
        self.verify_exception_with_msg(
            lambda: parse_cpu_quota_v2("200000 -1"),
            "Period in cgroups v2 cpu.max file was <= 0. Content: 200000 -1",
        )

    def test_get_cpu_used(self):
        t1 = monotonic()
        cu1 = get_cpu_used()
        x = 0
        for _ in range(10):
            for _ in range(100000):
                x += random()

            t2 = monotonic()
            cu2 = get_cpu_used()

            time_elapsed = t2 - t1
            cpu_used = cu2 - cu1
            # We're very lenient here, since tests could be running on an overbooked
            # and not have access to the full CPU
            assert cpu_used > time_elapsed * 0.5, "Expected test to consume > 50% CPU"
            (t1, cu1) = (t2, cu2)

    def verify_exception_with_msg(self, func, msg):
        with self.assertRaises(Exception) as ctx:
            func()
        assert str(ctx.exception) == msg



from collections import deque
from logging import getLogger
from os import environ, urandom
from threading import Lock, Thread, Timer
from time import monotonic, sleep
from typing import Any, Dict, Optional, Sequence, Tuple, Generic, TypeVar, Callable
from warnings import warn

from werkzeug.serving import make_server

from .wsgi_typing import (
    Environ,
    ExcInfo,
    ResponseStream,
    StartResponse,
    WsgiApplication,
)
from .util import ThreadsTracker

# Interval of how often to attempt cleanup of unclaimed jobs
CLEANUP_INTERVAL_SECONDS = float(environ.get("OOB_CLEANUP_INTERVAL_SECONDS", "5"))

# Maximum time a process is allowed to remain unclaimed before attempting to clean it up
ITEM_EXPIRY_SECONDS = float(environ.get("OOB_ITEM_EXPIRY_SECONDS", "20"))

LOGGER = getLogger("pothead.oob_response")


class Response:
    def __init__(self, headers, body, on_done):
        self.headers = headers
        self.body = body
        self.on_done = on_done

    def close(self):
        if self.on_done:
            self.body.close()
            self.on_done()
            self.on_done = None

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_value, trace):
        self.close()


class KeyedTransfer:
    """Allows one thread to push a value, and get a (url-safe string) key back.
    At a later time, another thread can retrieve the value by the key.

    With regular intervals, unclaimed keys are purged. If the value has a method
    `close()`, it is called on purging."""

    def __init__(self, cleanup_interval_seconds: float, item_expiry_seconds: float):
        self._cleanup_interval_seconds = cleanup_interval_seconds
        self._item_expiry_seconds = item_expiry_seconds
        self._dict: Dict[str, Tuple[float, Any]] = dict()
        self._lock = Lock()
        self._shutdown = False

        self._cleanup()

    def push(self, value: Any):
        key = urandom(24).hex()
        now = monotonic()
        expires_at = now + self._item_expiry_seconds
        with self._lock:
            assert key not in self._dict
            self._dict[key] = (expires_at, value)

        return key

    def pop(self, key) -> Optional[Any]:
        with self._lock:
            return self._dict.pop(key, (None, None))[1]

    def close(self):
        self._shutdown = True
        with self._lock:
            for (expires_at, job) in self._dict.values():
                if hasattr(job, "close"):
                    job.close()
            self._dict.clear()

    def wait_for_clear(self):
        while True:
            with self._lock:
                if len(self._dict) == 0:
                    return
            sleep(self._cleanup_interval_seconds / 2)

    def _cleanup(self):
        if not self._shutdown:
            t = Timer(self._cleanup_interval_seconds, self._cleanup)
            t.daemon = True
            t.start()
        now = monotonic()
        with self._lock:
            new_dict = dict()
            for key, (expires_at, job) in self._dict.items():
                if expires_at > now:
                    new_dict[key] = (expires_at, job)
                elif hasattr(job, "close"):
                    job.close()
            self._dict = new_dict


class PeekableIterator(object):
    def __init__(self, inner):
        self._inner = iter(inner)
        self._cache = deque()

    def __iter__(self):
        return self

    def __next__(self):
        if len(self._cache):
            return self._cache.popleft()
        else:
            return next(self._inner)

    def peek_next(self):
        if not len(self._cache):
            return self._cache.append(next(self._inner))
        return self._cache[0]

    def close(self):
        getattr(self._inner, "close", lambda: None)()


AtomicReferenceT = TypeVar("AtomicReferenceT")


class AtomicReference(Generic[AtomicReferenceT]):
    def __init__(self):
        self._lock = Lock()

    def is_set(self) -> bool:
        with self._lock:
            return hasattr(self, "data")

    def set(self, data: AtomicReferenceT) -> None:
        with self._lock:
            self.data = data

    def get(self) -> AtomicReferenceT:
        with self._lock:
            return self.data


class OutOfBandResponder:
    def __init__(
        self,
        wrapped_app: WsgiApplication,
        port=0,
        cleanup_interval_seconds: Optional[float] = None,
        item_expiry_seconds: Optional[float] = None,
    ):
        if cleanup_interval_seconds is None:
            cleanup_interval_seconds = CLEANUP_INTERVAL_SECONDS
        if item_expiry_seconds is None:
            item_expiry_seconds = ITEM_EXPIRY_SECONDS

        self._job_transfer = KeyedTransfer(
            cleanup_interval_seconds, item_expiry_seconds
        )
        self.wrapped_app = wrapped_app

        wsgi_app = self._content_response_wsgi
        self._server = make_server("0.0.0.0", port=port, app=wsgi_app, threaded=True)
        responder_service = Thread(target=self._server.serve_forever)
        responder_service.daemon = True
        responder_service.start()
        self._worker_tracker = ThreadsTracker()
        self._service_port = self._server.port

    def _content_response_wsgi(self, environ, start_response):
        with self._worker_tracker:
            path = environ["PATH_INFO"]
            if len(path) < 2 or path[0] != "/":
                start_response("404 Not Found", {})
                return []

            response_key = path[1:]
            response = self._job_transfer.pop(response_key)
            if response is None:
                start_response(
                    f"410 {response_key} is claimed, expired, or never registered", {}
                )
                return []

            with response:
                start_response(
                    "200 OK",
                    response.headers,
                )
                yield b""  # Flush response to client
                for chunk in response.body:
                    yield chunk

    def shutdown(self):
        LOGGER.info("Draining pending redirects")
        self._job_transfer.wait_for_clear()
        self._server.shutdown()
        LOGGER.info("Waiting for jobs to complete")
        self._worker_tracker.drain()
        LOGGER.info("Shutdown complete")

    def __call__(
        self,
        environ: Environ,
        start_response: StartResponse,
        assume_cleanup: Callable[[], Callable[[], None]],
    ) -> ResponseStream:
        start_response_called: AtomicReference[
            Tuple[str, Sequence[Tuple[str, str]], Optional[ExcInfo]]
        ] = AtomicReference()

        def my_start_response(
            status: str,
            headers: Sequence[Tuple[str, str]],
            exc_info: Optional[ExcInfo] = None,
        ):
            if start_response_called.is_set():
                warn("start_response called more than once")
            else:
                start_response_called.set((status, headers, exc_info))

        response = PeekableIterator(self.wrapped_app(environ, my_start_response))

        while not start_response_called.is_set():
            try:
                response.peek_next()
            except StopIteration:
                if start_response_called.is_set():
                    break
                else:
                    msg = "Application Terminated Without Response"
                    LOGGER.error(msg)
                    start_response("500 Internal Error", (), None)
                    return [msg.encode()]

        (status, headers, exc_info) = start_response_called.get()

        if not exc_info and status.startswith("200 "):
            on_done = assume_cleanup()
            key = self._job_transfer.push(Response(headers, response, on_done))
            start_response(
                "303 See Other",
                (
                    (
                        "Location",
                        f"http://{environ['SERVER_NAME']}:{self._service_port}/{key}",
                    ),
                ),
                None,
            )
            return []
        else:
            start_response(status, headers, exc_info)
            return response



from socket import AF_INET, SOCK_STREAM, socket, timeout as TimeoutError, SHUT_RDWR
from threading import Thread
from time import sleep, monotonic
from typing import Optional
from unittest import TestCase
from unittest.mock import MagicMock
from urllib.request import urlopen
from logging import INFO

from pytest import fixture

OOB_CLEANUP_INTERVAL_SECONDS = 0.05
OOB_ITEM_EXPIRY_SECONDS = 0.5
REQUEST_ID = "im-quite-unique!"

from pothead.gating import wait_for_idle_cpus  # noqa: E402 override environ first

from .util import ObjectProxy  # noqa: E402
from .worker import LoadBalancer, Server  # noqa: E402


def demo_app(environ, start_response):
    path = environ.get("PATH_INFO")
    if path == "/":
        start_response("200 OK", ())
        yield b"Hello World!"
    elif path == "/crash":
        raise Exception("Nope nope nope")
    elif path == "/slow_loris":
        sleep(0.05)
        start_response("200 OK", ())
        for byte in b"This might take a while...":
            sleep(0.1)
            yield byte
    else:
        start_response("404 Not Found", ())


class WorkerConnection:
    def __init__(self, sock: socket):
        self.sock = sock

    def send_request(self, path="/"):
        self.sock.sendall(
            f"GET {path} HTTP/1.1\r\nx-request-id: {REQUEST_ID}\r\n\r\n".encode("ascii")
        )

    def read_response(self):
        buf = b""
        read = True
        while read:
            read = self.sock.recv(1024)
            buf += read
        return buf

    def disconnect(self):
        self.sock.shutdown(SHUT_RDWR)
        self.sock.close()


class DummyBroker:
    def __init__(self):
        self.sock = socket(AF_INET, SOCK_STREAM)
        self.sock.bind(("localhost", 0))
        self.sock.listen(1)

    def addr(self):
        return self.sock.getsockname()

    def accept(self, timeout) -> Optional[WorkerConnection]:
        self.sock.settimeout(timeout)
        try:
            sock, _ = self.sock.accept()
            return WorkerConnection(sock)
        except TimeoutError:
            return None


def mock_balancer(addresses):
    class MockBalancer(LoadBalancer):
        def refresh(self):
            self._balancer.provision(addresses)

    return MockBalancer


class WorkerTest(TestCase):
    def test_load_balancing(self):
        broker_a, broker_b = (DummyBroker() for _ in range(2))
        app = ObjectProxy(demo_app)
        app.wait_for_slot = wait_for_idle_cpus(0, max_concurrent=10)

        worker = Server(
            ("", 0),
            app,
            load_balancer=mock_balancer(map(DummyBroker.addr, (broker_a, broker_b))),
            oob_cleanup_interval_seconds=OOB_CLEANUP_INTERVAL_SECONDS,
            oob_item_expiry_seconds=OOB_ITEM_EXPIRY_SECONDS,
        )
        Thread(target=worker.poll_loop, daemon=True).start()

        connection_a = broker_a.accept(0.1)
        connection_b = broker_b.accept(0.1)

        assert broker_a.accept(0.2) is None
        assert broker_b.accept(0.2) is None

        connection_b.send_request()
        assert broker_b.accept(0.2) is not None

        connection_a.send_request()
        assert broker_a.accept(0.2) is not None


class WaitForFirstSlot:
    def __init__(self, callbacks):
        self.counter = 0
        self.callbacks = callbacks

    def __call__(self, halt):
        self.counter += 1
        if self.counter <= 1:
            return self.callbacks
        else:
            from time import sleep

            while not halt():
                sleep(0.05)


class RedirectWorkerTest(TestCase):
    @fixture(autouse=True)
    def use_caplog(self, caplog):
        caplog.set_level(INFO)
        self.caplog = caplog

    def setUp(self):
        broker = DummyBroker()
        self.callbacks = MagicMock()

        app = ObjectProxy(demo_app)
        app.wait_for_slot = WaitForFirstSlot(self.callbacks)

        self.server = Server(
            ("", 0),
            app,
            load_balancer=mock_balancer([broker.addr()]),
            redirect_port=0,
            oob_cleanup_interval_seconds=OOB_CLEANUP_INTERVAL_SECONDS,
            oob_item_expiry_seconds=OOB_ITEM_EXPIRY_SECONDS,
        )
        Thread(target=self.server.poll_loop, daemon=True).start()
        self.connection = broker.accept(0.1)

    def tearDown(self):
        self.server.shutdown()

    def test_response_redirect(self):
        self.connection.send_request()
        resp = self.connection.read_response().split(b"\r\n")

        assert resp[0] == b"HTTP/1.1 303 See Other"
        location = (
            [x for x in resp if x.startswith(b"Location: ")][0].split()[1].decode()
        )
        real_response = urlopen(location)
        assert real_response.status == 200
        assert real_response.read().startswith(b"Hello World!")

    def test_reject_no_redirect(self):
        self.connection.send_request("/nonexisting")
        resp = self.connection.read_response().split(b"\r\n")

        assert resp[0] == b"HTTP/1.1 404 Not Found"
        assert b"Location" not in [x.split(b":") for x in resp]

    def test_fail_no_redirect(self):
        self.connection.send_request("/crash")
        resp = self.connection.read_response().split(b"\r\n")

        assert resp[0] == b"HTTP/1.1 500 INTERNAL SERVER ERROR"
        assert b"Location" not in [x.split(b":") for x in resp]

    def test_ondone_deferred(self):
        self.connection.send_request()
        resp = self.connection.read_response().split(b"\r\n")

        self.callbacks.done.assert_not_called()

        location = (
            [x for x in resp if x.startswith(b"Location: ")][0].split()[1].decode()
        )
        real_response = urlopen(location)
        assert real_response.status == 200
        assert real_response.read().startswith(b"Hello World!")

        self.callbacks.done.assert_called()

    def test_ondone_called_for_crash(self):
        self.connection.send_request("/crash")
        self.connection.read_response()

        self.callbacks.done.assert_called()

    def test_ondone_called_for_reject(self):
        self.connection.send_request("/nonexisting")
        self.connection.read_response()

        self.callbacks.done.assert_called()

    def test_ondone_called_after_timeout(self):
        self.connection.send_request()
        self.connection.read_response()

        self.callbacks.done.assert_not_called()

        sleep(OOB_ITEM_EXPIRY_SECONDS + 2 * OOB_CLEANUP_INTERVAL_SECONDS)

        self.callbacks.done.assert_called()

    def test_log_elapsed_after_success(self):
        self.connection.send_request()
        self.connection.read_response()

        sleep(0.05)

        assert f"[{REQUEST_ID}] Elapsed: " in self.caplog.text

    def test_log_elapsed_after_crash(self):
        self.connection.send_request("/crash")
        self.connection.read_response()

        sleep(0.05)

        assert f"[{REQUEST_ID}] Elapsed: " in self.caplog.text

    def test_log_elapsed_after_disconnect(self):
        self.connection.send_request("/slow_loris")
        self.connection.disconnect()

        sleep(0.2)

        assert f"[{REQUEST_ID}] Elapsed: " in self.caplog.text


class TestShutdown(TestCase):
    DELAY_START_RESPONSE = 0.3
    DELAY_FINISH_RESPONSE = 0.8

    def setUp(self):
        broker = DummyBroker()
        self.callbacks = MagicMock()

        def app(environ, start_response):
            sleep(self.DELAY_START_RESPONSE)
            start_response("200 OK", {})
            yield b" "
            sleep(self.DELAY_FINISH_RESPONSE)
            yield b"done"

        app.wait_for_slot = WaitForFirstSlot(self.callbacks)

        self.server = Server(
            ("", 0),
            app,
            load_balancer=mock_balancer([broker.addr()]),
            redirect_port=0,
            oob_cleanup_interval_seconds=OOB_CLEANUP_INTERVAL_SECONDS,
            oob_item_expiry_seconds=OOB_ITEM_EXPIRY_SECONDS,
        )
        Thread(target=self.server.poll_loop, daemon=True).start()
        self.connection = broker.accept(0.1)

    def test_ptth_finish_before_shutdown(self):
        self.connection.send_request()
        Thread(target=self.connection.read_response).start()
        self.assertGreaterEqual(
            self.shutdown_time(), OOB_ITEM_EXPIRY_SECONDS + self.DELAY_START_RESPONSE
        )

    def test_pending_oob_expire_before_shutdown(self):
        self.connection.send_request()
        self.connection.read_response()
        st = self.shutdown_time()
        self.assertGreaterEqual(st, OOB_ITEM_EXPIRY_SECONDS)
        self.assertLessEqual(st, self.DELAY_FINISH_RESPONSE)

    def test_oob_finish_before_shutdown(self):
        self.connection.send_request()
        resp = self.connection.read_response().split(b"\r\n")

        location = (
            [x for x in resp if x.startswith(b"Location: ")][0].split()[1].decode()
        )
        real_response = urlopen(location)
        assert real_response.status == 200

        self.assertGreaterEqual(self.shutdown_time(), self.DELAY_FINISH_RESPONSE)

    def shutdown_time(self):
        start = monotonic()
        self.server.shutdown()
        return monotonic() - start



import io
import os
import subprocess
from logging import getLogger
from psutil import Process
from threading import Thread
import pytest
import signal
import sys
import time
import timeout_decorator
import traceback
from typing import (
    Any,
    Callable,
    cast,
    Dict,
    Iterable,
    Optional,
    Sequence,
    TextIO,
    Tuple,
)
from unittest import TestCase
import werkzeug
from werkzeug.datastructures import Headers
from werkzeug.test import Client
from werkzeug.wsgi import LimitedStream

from .subprocess_middleware import SubprocessMiddleware
from .wsgi_typing import (
    Environ,
    ExcInfo,
    StartResponse,
    WriteCallable,
)

# Used for testing that only gets initialized once in the forkserver
# process, and not once per spawned subprocess.
TEST_GLOBAL_VAR = os.getpid()

LOGGER = getLogger("pothead.test_subprocess_middleware")


def assertActiveException() -> ExcInfo:
    exc_info = sys.exc_info()
    assert exc_info[0] is not None
    assert exc_info[1] is not None
    assert exc_info[2] is not None
    return exc_info


# This function executes inside the spawned subprocess
def wrapped_app(environ: Environ, start_response: StartResponse) -> Iterable[bytes]:
    # Separate code path for tests that do not read request payload
    if environ["PATH_INFO"] == "/test_success_app_does_not_read_payload":
        start_response("200 OK", [], None)
        yield b"This is a successful response"
    else:
        request = werkzeug.Request(cast(Dict[str, Any], environ))
        assert request.is_multiprocess

        if request.path == "/test_success_with_request_and_response_payloads":
            assert request.data == b"Non-UTF8 binary payload: \x80"
            start_response("200 OK", [("key1", "value1"), ("key2", "value2")], None)
            yield b"This is a "
            yield b"non-UTF8 success response: \x80"
        elif request.path == "/test_success_with_payload_and_cpu_reservation_reset":
            start_response("200 OK", [], None)
            assert "RESET_CPU_RESERVATION" in environ
            reset_cpu_reservation = cast(
                Callable[[int], None], environ["RESET_CPU_RESERVATION"]
            )
            yield b"This is a "
            reset_cpu_reservation(7)
            yield b"successful response"
        elif (
            request.path == "/test_no_cpu_reservation_reset_if_none_in_original_request"
        ):
            assert "RESET_CPU_RESERVATION" not in environ
            start_response("200 OK", [], None)
            yield b"This is a successful response"
        elif request.path == "/test_success_with_no_response_payload":
            start_response("200 OK", [], None)
        elif request.path == "/test_failed_request_with_wsgi_error_output":
            wsgi_errors = cast(TextIO, environ.get("wsgi.errors"))
            wsgi_errors.write("First error output\n")
            start_response("500 Internal Server Error", [], None)
            wsgi_errors.write("Second error output\n")
        elif request.path == "/test_exception_raised_in_wrapped_app":

            def a_function_in_the_wrapped_app_callstack():
                raise Exception("This is an exception message")

            a_function_in_the_wrapped_app_callstack()
        elif request.path == "/test_exception_returned_from_start_response":

            def a_function_in_the_wrapped_app_callstack():
                try:
                    raise Exception("This is an exception message")
                except Exception:
                    exc_info = assertActiveException()
                    start_response(
                        "500 Internal Server Error",
                        [("key1", "value1"), ("key2", "value2")],
                        exc_info,
                    )

            a_function_in_the_wrapped_app_callstack()
        elif request.path == "/test_wrapped_app_dies_from_signal":
            os.kill(os.getpid(), signal.SIGKILL)
        elif request.path == "/test_globals_not_reinitialized_for_each_request":
            start_response("200 OK", [], None)
            yield f"TEST_GLOBAL_VAR: {TEST_GLOBAL_VAR}".encode("utf-8")
        elif request.path in [
            "/test_successful_subprocess_is_shut_down",
            "/test_failing_subprocess_is_shut_down",
        ]:

            def sleep_forever():
                while True:
                    time.sleep(1)

            # This thread will prevent the subprocess from exiting normally
            t = Thread(target=sleep_forever, daemon=False)
            t.start()

            # This subprocess represents something that the app might spawn, that doesn't automatically exit on its own
            subprocess.Popen(
                ["bash", "-c", "while true; do sleep 1000; echo HI FROM APP; done"]
            )

            if request.path == "/test_successful_subprocess_is_shut_down":
                start_response("200 OK", [], None)
            else:
                raise Exception("This is an exception message")
        elif request.path == "/test_client_disconnect_while_sending_request":
            raise Exception("Should never make it to this point")
        else:
            assert False, f"Unknown test endpoint path: {request.path}"


class SubprocessMiddlewareTest(TestCase):
    def test_success_with_request_and_response_payloads(self):
        client = Client(SubprocessMiddleware(LOGGER, wrapped_app))
        response = client.post(
            "/test_success_with_request_and_response_payloads",
            data=b"Non-UTF8 binary payload: \x80",
        )

        assert response.status == "200 OK"
        assert response.headers == Headers(
            [
                ("key1", "value1"),
                ("key2", "value2"),
            ]
        )
        assert [b for b in response.response] == [
            b"This is a ",
            b"non-UTF8 success response: \x80",
        ]

    def test_success_app_does_not_read_payload(self):
        client = Client(SubprocessMiddleware(LOGGER, wrapped_app))
        large_payload_data = bytearray(20 * 1024 * 1024)
        response = client.post(
            "/test_success_app_does_not_read_payload",
            data=bytes(large_payload_data),
        )

        assert response.status == "200 OK"
        assert [b for b in response.response] == [b"This is a successful response"]

    def test_success_with_payload_and_cpu_reservation_reset(self):
        reset_cpu_calls = []

        def reset_cpu_reservation_callback(num_cpus: int):
            reset_cpu_calls.append(num_cpus)

        client = Client(SubprocessMiddleware(LOGGER, wrapped_app))
        response = client.post(
            "/test_success_with_payload_and_cpu_reservation_reset",
            environ_overrides={"RESET_CPU_RESERVATION": reset_cpu_reservation_callback},
        )

        assert response.status == "200 OK"
        assert [b for b in response.response] == [b"This is a ", b"successful response"]
        assert reset_cpu_calls == [7]

    def test_no_cpu_reservation_reset_if_none_in_original_request(self):
        client = Client(SubprocessMiddleware(LOGGER, wrapped_app))
        response = client.post(
            "/test_no_cpu_reservation_reset_if_none_in_original_request",
        )

        assert response.status == "200 OK"
        assert [b for b in response.response] == [b"This is a successful response"]

    def test_failed_request_with_wsgi_error_output(self):
        client = Client(SubprocessMiddleware(LOGGER, wrapped_app))

        error_output = io.StringIO()
        response = client.post(
            "/test_failed_request_with_wsgi_error_output",
            environ_overrides={"wsgi.errors": error_output},
        )

        assert response.status == "500 Internal Server Error"
        assert [b for b in response.response] == []
        assert error_output.getvalue() == "First error output\nSecond error output\n"

    def test_exception_raised_in_wrapped_app(self):
        client = Client(SubprocessMiddleware(LOGGER, wrapped_app))

        with pytest.raises(Exception) as pytest_exc_info:
            client.post("/test_exception_raised_in_wrapped_app")

        exc_msg = pytest_exc_info.value.args[0]
        exc_traceback = "\n".join(traceback.format_tb(pytest_exc_info.tb))
        assert "This is an exception message" in exc_msg
        assert "test_subprocess_middleware.py" in exc_traceback
        assert "a_function_in_the_wrapped_app_callstack" in exc_traceback

    def test_exception_returned_from_start_response(self):
        client = Client(SubprocessMiddleware(LOGGER, wrapped_app))

        with pytest.raises(Exception) as pytest_exc_info:
            client.post("/test_exception_returned_from_start_response")

        exc_msg = pytest_exc_info.value.args[0]
        exc_traceback = "\n".join(traceback.format_tb(pytest_exc_info.tb))
        assert "This is an exception message" in exc_msg
        assert "test_subprocess_middleware.py" in exc_traceback
        assert "a_function_in_the_wrapped_app_callstack" in exc_traceback

    def test_wrapped_app_dies_from_signal(self):
        client = Client(SubprocessMiddleware(LOGGER, wrapped_app))

        with pytest.raises(Exception) as pytest_exc_info:
            client.post("/test_wrapped_app_dies_from_signal")

        exc_msg = pytest_exc_info.value.args[0]
        assert "Failed to read from WSGI request subprocess!" in exc_msg

    def test_globals_not_reinitialized_for_each_request(self):
        client = Client(SubprocessMiddleware(LOGGER, wrapped_app))

        response1 = client.post("/test_globals_not_reinitialized_for_each_request")
        response2 = client.post("/test_globals_not_reinitialized_for_each_request")

        response1_data = [b for b in response1.response]
        response2_data = [b for b in response2.response]

        assert response1_data[0].startswith(b"TEST_GLOBAL_VAR:")
        assert response1_data == response2_data

    def test_successful_subprocess_is_shut_down(self):
        exc = self.do_subprocess_shutdown_test(
            lambda middleware: Client(middleware).post(
                "/test_successful_subprocess_is_shut_down"
            )
        )
        assert exc is None

    def test_failing_subprocess_is_shut_down(self):
        exc = self.do_subprocess_shutdown_test(
            lambda middleware: Client(middleware).post(
                "/test_failing_subprocess_is_shut_down"
            )
        )
        assert exc is not None
        assert "This is an exception message" in exc.args[0]

    def test_client_disconnect_while_sending_request(self):
        def do_request(middleware: SubprocessMiddleware):
            # We do a manual WSGI call here, rather than using client.post(), so
            # that we can directly inject a LimitedStream into the middleware
            # that will throw a ClientDisconnected exception when reading from
            # it.
            # If we used client.post(), the exception would instead happen
            # outside of the middleware, which doesn't match what will happen in
            # real life if the client disconnects.
            environ = {
                "PATH_INFO": "/test_client_disconnect_while_sending_request",
                "CONTENT_LENGTH": 1,
                "wsgi.input": LimitedStream(io.BytesIO(), 1),
                "wsgi.errors": None,
            }

            def start_response(
                status: str,
                headers: Sequence[Tuple[str, str]],
                exc_info: Optional[ExcInfo],
            ) -> WriteCallable:
                assert False, "start_response callback should not have been called"

            for b in middleware(environ, start_response):
                pass

        exc = self.do_subprocess_shutdown_test(do_request)
        assert exc is not None
        assert "Client disconnected while writing request payload" in exc.args[0]

    @timeout_decorator.timeout(2)
    def do_subprocess_shutdown_test(
        self, request_func: Callable[[SubprocessMiddleware], None]
    ) -> Optional[Exception]:
        middleware = SubprocessMiddleware(LOGGER, wrapped_app)

        proc = Process(os.getpid())

        init_process = proc
        while init_process.pid != 1:
            init_process = init_process.parent()

        subprocesses_before = set(proc.children(recursive=True))
        init_subprocesses_before = set(init_process.children(recursive=True))

        try:
            request_func(middleware)
            return None
        except timeout_decorator.TimeoutError:
            raise
        except Exception as exc:
            return exc
        finally:
            subprocesses_after = set(proc.children(recursive=True))
            if subprocesses_before != subprocesses_after:
                # Avoid hanging the test itself in case it fails
                extra_processes = subprocesses_after.difference(subprocesses_before)
                for p in extra_processes:
                    p.terminate()

                assert False, f"Request subprocess didn't exit: {extra_processes}"

            # If the app leaked running subprocesses upon being killed, these will be orphaned by
            # the init process. To avoid getting false positives on this (in case the init process has
            # adopted some unrelated subprocesses while the test was being run), we only conclude that
            # a process was leaked if its cmdline matches a known string.
            init_subprocesses_after = set(init_process.children(recursive=True))
            if init_subprocesses_before != init_subprocesses_after:
                # Avoid hanging the test itself in case it fails
                extra_processes = init_subprocesses_after.difference(
                    init_subprocesses_before
                )
                leaked_subprocess = None
                for p in extra_processes:
                    if any(word for word in p.cmdline() if "HI FROM APP" in word):
                        leaked_subprocess = p
                        print(
                            f"Leaked subprocess: {leaked_subprocess} ({leaked_subprocess.cmdline()})"
                        )
                    p.terminate()

                if leaked_subprocess:
                    assert False, f"Subprocess was leaked: {leaked_subprocess}"



"""
WSGI middleware that spawns off a separate subprocess for each request.

Note: this is currently only suitable for use cases where the size of the
request payload is small, since it is read up-front instead of being streamed to
the inner WSGI app.
"""

import io
import multiprocessing
import psutil
from logging import Logger
from multiprocessing.connection import Connection
from tblib import pickling_support
from threading import Thread
import werkzeug

from typing import (
    cast,
    Any,
    Callable,
    Dict,
    Iterable,
    NamedTuple,
    Optional,
    Sequence,
    TextIO,
    Tuple,
)
from .wsgi_typing import (
    Environ,
    ExcInfo,
    StartResponse,
    WriteCallable,
    WsgiApplication,
)

# So that exception tracebacks can be moved across process boundaries
pickling_support.install()


class MsgStartResponse(NamedTuple):
    """
    Sent when the wrapped WSGI app has invoked the `start_response` callback.
    """

    http_status: str
    response_headers: Sequence[Tuple[str, str]]
    exc_info: Optional[ExcInfo]


class MsgData(NamedTuple):
    """
    Sent when the wrapped WSGI app generates payload bytes via `yield`.
    """

    data: bytes


class MsgWriteErrors(NamedTuple):
    """
    Sent when the wrapped WSGI app writes to "wsgi.errors" in its environment.
    """

    data: str


class MsgResetCpuReservation(NamedTuple):
    """
    Sent when the wrapped WSGI app invokes the RESET_CPU_RESERVATION callback
    (from the environment).
    """

    num_cpus: int


class MsgExceptionRaised(NamedTuple):
    """
    Sent if the wrapped WSGI app raises an exception while processing the request.
    """

    exc: BaseException


class MsgDone:
    """
    Sent when the wrapped WSGI app has finished processing the request without
    raising an exception. Note that this does not necessarily mean that the HTTP
    request itself was successful.
    """

    pass


def do_nothing_warmup():
    pass


class SubprocessMiddleware:
    def __init__(self, logger: Logger, wrapped_app: WsgiApplication):
        """
        wrapped_app: WsgiApplication
            The underlying WSGI app, which will have each of its requests called
            in a separate subprocess.
        """
        self.logger = logger
        self.wrapped_app = wrapped_app

        # Use a forkserver with the app's module preloaded, to make the
        # spawning of subprocesses for each request faster.
        self.forkserver_context = multiprocessing.get_context("forkserver")
        preload_module_names = [wrapped_app.__module__]
        self.forkserver_context.set_forkserver_preload(preload_module_names)

        # Spawning the very first subprocess instance is a lot slower than subsequent
        # spawns, so do that right away to make the actual requests fast.
        warmup_subprocess = self.forkserver_context.Process(target=do_nothing_warmup)
        warmup_subprocess.start()
        warmup_subprocess.join()

    def __call__(
        self, environ: Environ, start_response: StartResponse
    ) -> Iterable[bytes]:
        """
        Implementation of the WSGI app interface.
        """

        new_environ = {k: v for k, v in environ.items()}
        new_environ["wsgi.multiprocess"] = True
        new_environ["wsgi.multithread"] = False
        new_environ.pop("werkzeug.request", None)
        new_environ.pop("werkzeug.server.shutdown", None)

        # These two streams are replaced by pipe communication with the subprocess
        new_environ.pop("wsgi.input")
        wsgi_errors = cast(TextIO, new_environ.pop("wsgi.errors"))

        # RESET_CPU_RESERVATION is a callback function, which cannot be moved
        # across process boundaries. The provided callback (if any) is saved
        # here, and then invoked when the subprocess sends a specific message
        # over the pipe.
        reset_cpu_reservation = cast(
            Optional[Callable[[int], None]],
            new_environ.pop("RESET_CPU_RESERVATION", None),
        )

        # Set up the thread for reading the wsgi.input data
        (wsgi_input_reader, wsgi_input_writer) = self.forkserver_context.Pipe(
            duplex=False
        )
        request_sender_thread = Thread(
            target=SubprocessMiddleware.send_request_data,
            args=(self.logger, environ, wsgi_input_writer),
        )

        # Setup up the subprocess for running the wrapped WSGI app
        (msg_reader, msg_writer) = self.forkserver_context.Pipe(duplex=False)
        subprocess = self.forkserver_context.Process(
            target=subprocess_main,
            name="WSGI request subprocess",
            args=(
                self.wrapped_app,
                new_environ,
                msg_writer,
                wsgi_input_reader,
                reset_cpu_reservation is not None,
            ),
        )
        try:
            request_sender_thread.start()
            subprocess.start()

            # Close our instances of the unused sides of the connections, so that
            # once the subprocess closes its instance, the connection will be
            # completely closed and recv() and similar methods will indicate EOF.
            msg_writer.close()
            wsgi_input_reader.close()

            while True:
                try:
                    msg = msg_reader.recv()
                except EOFError:
                    # This should not normally happen even if the wrapped app raises
                    # an exception during processing, and likely indicates that the
                    # subprocess died from receiving a signal.
                    raise Exception("Failed to read from WSGI request subprocess!")

                if isinstance(msg, MsgStartResponse):
                    start_response(msg.http_status, msg.response_headers, msg.exc_info)
                elif isinstance(msg, MsgData):
                    yield msg.data
                elif isinstance(msg, MsgWriteErrors):
                    wsgi_errors.write(msg.data)
                elif isinstance(msg, MsgResetCpuReservation):
                    if reset_cpu_reservation is not None:
                        reset_cpu_reservation(msg.num_cpus)
                elif isinstance(msg, MsgExceptionRaised):
                    raise msg.exc
                elif isinstance(msg, MsgDone):
                    break
                else:
                    assert (
                        False
                    ), f"Got unknown message type from WSGI subprocess: {msg}"
        finally:
            _kill_process_and_its_descendants(subprocess)
            request_sender_thread.join()

    @staticmethod
    def send_request_data(logger: Logger, environ: Environ, writer: Connection):
        try:
            with writer:
                request_data = werkzeug.Request(
                    cast(Dict[str, Any], environ)
                ).stream.read()
                writer.send_bytes(request_data)
        except werkzeug.exceptions.ClientDisconnected:
            # Since we have closed the writer at this point, the subprocess will
            # fail to read from the other side, and that exception will be
            # transported to the main process. So we can just log here, and then
            # let the thread finish.
            logger.warning("Request reader thread exited due to client disconnect")


# Sends data written by the wrapped app to wsgi.errors back to
# the main process via a pipe.
class WSGIErrors:
    def __init__(self, writer: Connection):
        self.writer = writer

    def write(self, data: str):
        self.writer.send(MsgWriteErrors(data))


def subprocess_main(
    wrapped_app: WsgiApplication,
    environ: Environ,
    msg_writer: Connection,
    wsgi_input_reader: Connection,
    has_reset_cpu_reservation_callback,
):
    with msg_writer, wsgi_input_reader:
        try:

            def wrapped_start_response(
                http_status: str,
                headers: Sequence[Tuple[str, str]],
                exc_info: Optional[ExcInfo] = None,
            ) -> WriteCallable:
                """Implements the "start response callback" WSGI interface"""

                msg_writer.send(MsgStartResponse(http_status, headers, exc_info))

                def disabled_write_callback(data: bytes) -> Any:
                    raise Exception(
                        "Write callback not supported with SubprocessMiddleware: "
                        + "Use `yield` to generate payload data instead."
                    )

                return disabled_write_callback

            def reset_cpu_reservation(num_cpus: int):
                msg_writer.send(MsgResetCpuReservation(num_cpus))

            new_environ = {k: v for k, v in environ.items()}

            # Set up the CPU reservation callback if the original request had one
            assert "RESET_CPU_RESERVATION" not in environ
            if has_reset_cpu_reservation_callback:
                new_environ["RESET_CPU_RESERVATION"] = reset_cpu_reservation

            # Read the request data from the parent process
            try:
                request_data = wsgi_input_reader.recv_bytes()
            except EOFError:
                raise Exception("Client disconnected while writing request payload")

            assert isinstance(request_data, bytes)
            new_environ["wsgi.input"] = io.BytesIO(request_data)

            # Set up a proxy output stream for WSGI errors that send them
            # back to the parent process.
            new_environ["wsgi.errors"] = WSGIErrors(msg_writer)

            # Make the actual WSGI call to the wrapped app
            for output_data in wrapped_app(new_environ, wrapped_start_response):
                msg_writer.send(MsgData(output_data))

            msg_writer.send(MsgDone())
        except Exception as exc:
            msg_writer.send(MsgExceptionRaised(exc))
            raise


def _suspend_recursive(pid: psutil.Process) -> Iterable[psutil.Process]:
    """Gets a list of process and descendants. The only way to do it ~"atomically", is to suspend
    them in the process, preventing them from forking or dying while we're looking the other way.

    Returns a list of now suspended processes, including the root specified by pid"""
    try:
        proc = psutil.Process(pid)
        proc.suspend()
    except psutil.NoSuchProcess:
        return []

    descendants = {proc}
    while True:
        prior_count = len(descendants)
        try:
            for child in proc.children(recursive=True):
                if child not in descendants:
                    child.suspend()
                descendants.add(child)
        except psutil.NoSuchProcess:
            if proc.is_running():
                continue
            else:
                return []
        if len(descendants) == prior_count:
            break

    return descendants


def _kill_process_and_its_descendants(process: multiprocessing.Process):
    descendants = _suspend_recursive(process.pid)

    for descendant in descendants:
        try:
            descendant.kill()
        except psutil.NoSuchProcess:
            # Without a defined order, descendants might be already orphaned, adopted and killed
            # by init when we get to them
            continue

    process.join()
